/*anomaly.c */

/* (c) N Grassly and A Rambaut 1998*/
/*Department of Zoology, Oxford University*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lik.h"
#include "calclik.h"
#include "anomaly.h"
#include "evolve.h"
#include "progress.h"

/*variables*/
int numReps;

double likelihood;
double *actualLik, *simLik;

double **winLikList, **liktable;
double *variance, *mean;
int half, half_plus_one, numWindowSizes;
int numReps, minWinSize, testRegions[20], numTestRegions;

/*prototypes*/

void Results();
static int Smooth();
int compare(const void *n1, const void *n2);
static void Actualsurf();
static void CalcMeanVar();
static void AllocWinLikMem();
static void DeAllocWinLikMem();
static double **AllocLikSurf(char *surfName);
static double normZval(double P, short *IFAULT);



/****************/
void Anomaly(TTree *tree)
{
int i, j , k, size, sp;
double temp;
char *Q, *P;

	likelihood=tree->lnL;
	half=(int) (numBases/2);
	half_plus_one=half+1;
	numWindowSizes=half_plus_one-minWinSize;
	
	actualLik=(double *)AllocMem(sizeof(double)*numBases, "actualLik", "Anomaly", 0);
	if (codonRates) {
		for(j=0;j<numBases;j++)
			actualLik[j]=log(patternLik[codonCats[j % 3]][patternIndex[j]]);
	} else {
		for(j=0;j<numBases;j++)
			actualLik[j]=log(patternLik[0][patternIndex[j]]);
	}
		
	simLik=(double *)AllocMem(sizeof(double)*numBases, "simLik", "Anomaly", 0);
	
	AllocWinLikMem();

	if (!quiet) {
		InitProgressBar(numReps);
		DrawProgressBar();
	}
	
	for(i=0;i<numReps;i++) {
		EvolveSequences(tree);
		for (j=0; j<numTaxa; j++) {
			Q=tree->nodes[j].sequence;
			P=sequences[j];
			for (k=0; k<numBases; k++) {
				*P=(*Q);
				P++;
				Q++;
			}
		}
		
		CreatePatterns();
		AllocateTreeLikelihoods(tree);
		SetTipLikelihoods(tree);
		likelihood=CalcLik();
		if (codonRates) {
			for(j=0;j<numBases;j++)
				simLik[j]=log(patternLik[codonCats[j % 3]][patternIndex[j]]);
		} else {
			for(j=0;j<numBases;j++)
				simLik[j]=log(patternLik[0][patternIndex[j]]);
		}
		
		for(size=minWinSize;size<half_plus_one;size++){
			for(sp=0;sp<(numBases+1-size);sp++){
				temp=0.0;			
				for(j=sp;j<(sp+size);j++)
					temp+=simLik[j];
				temp=(temp/size)/((likelihood-temp)/(numBases-size));
				if(temp>winLikList[size-minWinSize][i])
					winLikList[size-minWinSize][i]=temp;
			}
		}
		
		if (!quiet)
			ProgressBar();
	}
	
	CalcMeanVar();
	DeAllocWinLikMem();
	likelihood=tree->lnL;
	Results();
}

void Results()
{
short  check, *map;
int i, j, n, size, sp;
double temp, *P, alpha, Z_cutOff;
short dummy;
		
	alpha=0.05/numWindowSizes;
	liktable=AllocLikSurf("liktable");
	Actualsurf();			
	map=(short *)CAllocMem((sizeof(short)*numBases), "map", "Results", 1);
	for(i=0;i<numWindowSizes;i++){
		for(j=0;j<(numBases+1-i-minWinSize);j++)
			liktable[i][j]=(liktable[i][j]-mean[i])/(sqrt(variance[i]));
	}

	fprintf(stderr, "\nBonferroni-corrected significance level for alpha=0.05: %e\n", alpha);
	
	Z_cutOff=-normZval(alpha, &dummy);
	
	fprintf(stderr, "\tZ-values greater than %f are significant\n\n", Z_cutOff);
	if(verbose){
		fprintf(stderr, "Verbose mode selected: all anomalies for Z > 1.0 will be output\n\n");
		Z_cutOff=1.0;
	}
	temp=0.0;
	for(i=minWinSize;i<half_plus_one;i++){
		P=liktable[i-minWinSize];
		for(j=0;j<(numBases+1-i);j++){
			if(*P>temp && *P>Z_cutOff){	
				temp=*P;
				size=i;
				sp=j;
			}
			P++;
		}
	}
	if(temp>Z_cutOff){
		fprintf(stdout, "Start End Z-value\n");
		do{
			for(i=sp;i<(sp+size);i++)
				map[i]=1;
			fprintf(stdout, "%d\t%d\t%f\n", (sp+1), (sp+size), liktable[size-minWinSize][sp]);			
			temp=0.0;
			for(i=minWinSize;i<half_plus_one;i++){
				P=liktable[i-minWinSize];
				for(j=0;j<(numBases+1-i);j++){
					if(*P>temp && *P>2.58){
						check=0;
						for(n=j;n<(j+i);n++)
							check+=map[n];
						if(check==0){
							temp=*P;
							size=i;
							sp=j;
						}
					}
					P++;
				}
			}
		}while(temp>Z_cutOff);
	}else
		fprintf(stderr, "No anomalous regions found\n");
	
	if(numTestRegions){
		fprintf(stderr, "Z-values for user entered regions:\n\n");
			fprintf(stdout, "Start End Z-value\n");
		for(i=0;i<numTestRegions;i++){
			sp=testRegions[i*2]-1;
			size=testRegions[(i*2)+1]-testRegions[(i*2)]+1;
			if(size<minWinSize)
				fprintf(stderr, "Error: regions to be tested must be bigger than the MIN_WINDOW_SIZE\n");
			else
				fprintf(stdout, "%d\t%d\t%f\n", (sp+1), (sp+size), liktable[size-minWinSize][sp]);			
		}
	}
	
	fprintf(stderr, "\nFinished\n");
	
}	

void Actualsurf()
{
double temp;
int j, size, sp;

	for(size=minWinSize;size<half_plus_one;size++){
		for(sp=0;sp<(numBases+1-size);sp++){
			temp=0.0;			
			for(j=sp;j<(sp+size);j++)
				temp+=actualLik[j];
			temp=(temp/size)/((likelihood-temp)/(numBases-size));
			liktable[size-minWinSize][sp]=temp;
		}
	}
}

void CalcMeanVar()
{
int i, j;
long mem;
double sums, sumsq;

	mem=totalMem;
	variance=(double *)AllocMem((sizeof(double)*numWindowSizes), "variance", "Results", 0);
	mean=(double *)AllocMem((sizeof(double)*numWindowSizes), "mean", "Results", 0);
	if (verboseMemory) 
		fprintf(stderr, "MEM: Z-value calculations - %ld bytes\n", totalMem-mem);
		
	sumsq=sums=0.0;
	for(i=0;i<numWindowSizes;i++){
		sumsq=sums=0.0;
		for(j=0;j<numReps;j++){
			sums+=winLikList[i][j];
			sumsq+=(winLikList[i][j]*winLikList[i][j]);
		}
		variance[i]=((sumsq-((sums*sums)/numReps))/(numReps-1));
		mean[i]=sums/numReps;
	}

}

void AllocWinLikMem()
{
int i;
long mem;
	
	mem=totalMem;
	winLikList=(double **)AllocMem(sizeof(double *)*numWindowSizes, "winLikList", "AllocWinLikMem", 0);
	for(i=0;i<numWindowSizes;i++)
		winLikList[i]=(double *)CAllocMem((sizeof(double)*numReps), "winLikList[]", "AllocWinLikMem", 0);
	if (verboseMemory)
		fprintf(stderr, "MEM: Null likelihood surface arrays- %ld bytes\n", totalMem-mem);
}

void DeAllocWinLikMem()
{
int i;	
	for(i=0;i<numWindowSizes;i++)
		free(winLikList[i]);
	free(winLikList);
	if (verboseMemory)
		fprintf(stderr, "MEM: Null likelihood surface arrays freed\n");
}

double **AllocLikSurf(char *surfName)
{
int i;
long mem;
double **P;

	mem=totalMem;
	P=(double **)AllocMem((sizeof(double *)*numWindowSizes), surfName, "AllocLikSurf", 0);
	for (i=0;i<numWindowSizes;i++)
		P[i]=(double *)AllocMem(sizeof(double)*(numBases+1-i-minWinSize), surfName, "AllocLikSurf", 0);
	if (verboseMemory)
		fprintf(stderr, "MEM: sequence data likelihood surface - %ld bytes\n", totalMem-mem);
	return P;
}

double normZval(double P, short *IFAULT){
/*
	modified from Fortran algorithm AS241
	APPL. STATIST. (1988) VOL. 37, NO. 3, 477-484.

	Produces the normal deviate Z corresponding to a given lower
	tail area of P; Z is accurate to about 1 part in 10**7.
*/

	double  SPLIT1, SPLIT2, CONST1, CONST2, A0, A1,
     		A2, A3, B1, B2, B3, C0, C1, C2, C3, D1, D2, E0, E1, E2,
     		E3, F1, F2, Q, R;
	double x;
	
    SPLIT1 = 0.425;
    SPLIT2 = 5.0,
    CONST1 = 0.180625;
    CONST2 = 1.6;
/*
	Coefficients for P close to 0.5
*/
	A0 = 3.3871327179E+00;
	A1 = 5.0434271938E+01;
    A2 = 1.5929113202E+02;
    A3 = 5.9109374720E+01;
    B1 = 1.7895169469E+01;
    B2 = 7.8757757664E+01;
    B3 = 6.7187563600E+01;
    
/*
	Coefficients for P not close to 0, 0.5 or 1.
*/
	C0 = 1.4234372777E+00;
	C1 = 2.7568153900E+00;
    C2 = 1.3067284816E+00;
    C3 = 1.7023821103E-01;
    D1 = 7.3700164250E-01;
    D2 = 1.2021132975E-01;

/*
	Coefficients for P near 0 or 1.
*/
	E0 = 6.6579051150E+00;
	E1 = 3.0812263860E+00;
    E2 = 4.2868294337E-01;
    E3 = 1.7337203997E-02;
    F1 = 2.4197894225E-01;
    F2 = 1.2258202635E-02;

	*IFAULT = 0;
	Q = P - 0.5;
	if(fabs(Q) <= SPLIT1){
		R = CONST1 - (Q * Q);
		x= Q * (((A3 * R + A2) * R + A1) * R + A0) /	
			(((B3 * R + B2) * R + B1) * R + 1.0);
		return x;
	}else if(Q < 0.0)
		R = P;
	else
	    R = 1.0 - P;

	if(R<=0.0){
		*IFAULT = 1;
	    return 0.0;
	}
	
	R = sqrt(-log(R));
	
	if(R <= SPLIT2){
	    R = R - CONST2;
	    x = (((C3 * R + C2) * R + C1) * R + C0) /
			 ((D2 * R + D1) * R + 1.0);
	}else{
		R = R - SPLIT2;
		x = (((E3 * R + E2) * R + E1) * R + E0) /
			((F2 * R + F1) * R + 1.0);
	}
	if(Q<0.0)
		x = -x;

	return x;
}

