#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "lik.h"
#include "evolve.h"
#include "random.h"

char *nucTable="ACGT";


/* prototypes */

char SetBase(double *P);
char SetBaseAccum(double *P);
void RandomSequence(char *seq);
void MutateSequence(char *seq, int branch);
void EvolveNode(TNode *anc, TNode *des);


/* functions */


char SetBase(double *P)
{
	char j;
	double r;
	
	r=rndu();
	for (j='\0'; r>(*P); j++) P++;
	return j;
}


char SetBaseAccum(double *P)
{
	char j;
	double r, sum;
	
	r=rndu();
	sum=*P;
	for (j='\0'; r>sum; j++) {
		P++;
		sum+=*P;
	}
	return j;
}


void RandomSequence(char *seq)
{
	int i;
	char *P;
	
	P=seq;
	for (i=0; i<numBases; i++) {
		*P=SetBase(addFreq);
		P++;
	}
	*P='\0';
}


void MutateSequence(char *seq, int branch)
{
	int i, mat;
	double r, sum, *Q;
	short *T;
	char *P;
	
	P=seq;
	
	mat=branch<<4;
	T=categories;
	for (i=0; i<numBases; i++) {
		Q=transMatrix[*T] + mat + (*P<<2);
		
		r=rndu();
		sum=*Q;
		*P=0;
		while (r>sum) {
			Q++;
			sum+=*Q;
			(*P)++;
		}
		
		P++;
		T++;
	}
}


void EvolveNode(TNode *anc, TNode *des)
{
	if (des->sequence==NULL)
		des->sequence=(char *)AllocMem(sizeof(char)*numBases, "sequence", "EvolveSequences", 0);
		
	memcpy(des->sequence, anc->sequence, numBases);
	MutateSequence(des->sequence, des->branchNo);
	
	if (des->tipNo==-1) {
		EvolveNode(des, des->next->branch);
		EvolveNode(des, des->next->next->branch);
	}
}


void EvolveSequences(TTree *tree)
/* Evolves sequences down tree using its branch lengths and
   Probability Transition Matrix and the nucleotide frequencies */
{
	int i, j;
	char *Q;
	
	if (tree->start->sequence==NULL)
		tree->start->sequence=(char *)AllocMem(sizeof(char)*numBases, "sequence", "EvolveSequences", 0);
		
	if (gammaRates) {
		for (i=0; i<numBases; i++)
			categories[i]=floor(rndu()*numCategories);
	} else {
		for (i=0; i<numBases; i++)
			categories[i]=i%numCategories;
	}
	
	RandomSequence(tree->start->sequence);
	
	EvolveNode(tree->start, tree->start->branch);
	EvolveNode(tree->start, tree->start->next->branch);
	EvolveNode(tree->start, tree->start->next->next->branch);
	
	for (i=0; i<numTaxa; i++) {
		Q=tree->nodes[i].sequence;
		for (j=0; j<numBases; j++) {
			*Q=nucTable[*Q];
			Q++;
		}
	}
}
